# $Id: S5DAGGeneratorA.py,v 1.17 2005/09/20 18:49:49 kipp Exp $

#
# Preamble.
#

from glue import lal
from glue import segments
import onasys.DAGGenerator
from onasys import onasysd
import logging
from math import ceil, floor
import os
import popen2


#
# Helpful functions.
#

def round_down(l, min, inc):
	"""
	Compute the greatest integer i, such that
	0 <= i <= l, i = (min + inc n) for n E {0, 1, 2...}
	"""
	if l < min:
		return 0
	if not inc:
		return min
	return l - (l - min) % inc

def compute_boundaries(config, sciseg, analseg):
	"""
	Given configuration information, a science segment, and an unused
	analysis segment, compute the corrected boundaries of the analysis
	segment and the surrounding data segment.
	"""
	# Construct a preliminary data segment by extending the boundaries
	# of the analysis segment the amount of the transient minima.  If a
	# boundary extends beyond the science segment, truncate at the
	# science segment boundary.
	dataseg = segments.segment(max(sciseg[0], analseg[0] - config.transient_left_minimum), min(sciseg[1], analseg[1] + config.transient_right_minimum))

	# Check for minimum length.  This guarantees that the analysis
	# segment won't be adjusted to 0 or negative lengths in the next
	# two steps.
	if dataseg.duration() < config.analysis_length_minimum + config.transient_left_minimum + config.transient_right_minimum:
		return (None, None)

	# Now, do the inverse:  given the data segment boundaries, set the
	# analysis segment boundaries according to the transient minima.
	# If the data segment was truncated against a science segment
	# boundary, this corrects the corresponding analysis segment
	# boundary, otherwise it has no effect.
	analseg = segments.segment(dataseg[0] + config.transient_left_minimum, dataseg[1] - config.transient_right_minimum)

	# Adjust the length of the analysis segment down to the nearest
	# allowed length, and adjust the data segment too.
	analseg = segments.segment(analseg[0], analseg[0] + round_down(float(analseg.duration()), config.analysis_length_minimum, config.analysis_length_increment))
	dataseg = segments.segment(dataseg[0], analseg[1] + config.transient_right_minimum)

	# Check for minimum length.  This avoids divide-by-zero errors
	# unless we really don't have enough data and the transient
	# increments are 0.
	if dataseg.duration() >= config.data_length_minimum:
		return (dataseg, analseg)

	# Determine the number of increments of the data segment lower
	# boundary required to make up the data length minimum, and take as
	# many as possible.  If we don't need any, then we're done.
	i = ceil((config.data_length_minimum - float(dataseg.duration())) / config.transient_left_increment)
	if i <= 0.0:
		return (dataseg, analseg)
	dataseg = segments.segment(dataseg[0] - min(i, floor(float(dataseg[0] - sciseg[0]) / config.transient_left_increment)) * config.transient_left_increment, dataseg[1])

	# Determine the number of increments of the data segment upper
	# boundary required to make up the data length minimum, and take as
	# many as possible.  If we don't need any, then we're done.
	i = ceil((config.data_length_minimum - float(dataseg.duration())) / config.transient_right_increment)
	if i <= 0.0:
		return (dataseg, analseg)
	dataseg = segments.segment(dataseg[0], dataseg[1] + min(i, floor(float(sciseg[1] - dataseg[1]) / config.transient_right_increment)) * config.transient_right_increment)

	# Check for minimum length.
	if dataseg.duration() <= config.data_length_minimum:
		return (None, None)
	return (dataseg, analseg)


#
# Default DAG generator interface.
#

class DAGGeneratorConfig(onasys.DAGGenerator.DAGGeneratorConfig):
	def defaults(self):
		return [
			("data_dir", os.getcwd()),
			("extra_args", ""),
		]

	def parse(self):
		self.parser.read(self.filename)
		for key in ["data_dir", "extra_args", "generator"]:
			self.__dict__[key] = self.parser.get(self.section, key)
		for key in ["cusion_timeout", "data_length_minimum", "analysis_length_minimum", "analysis_length_increment", "transient_left_minimum", "transient_left_increment", "transient_right_minimum", "transient_right_increment"]:
			self.__dict__[key] = self.parser.getfloat(self.section, key)

	def validate(self):
		if self.data_length_minimum < self.analysis_length_minimum + self.transient_left_minimum + self.transient_right_minimum:
			raise ConfigError, ConfigErrorMsg(self, "sum of analysis_length, transient_left and transient_right minima excedes data_length_minimum")

class DAGGeneratorParameters(onasys.DAGGenerator.DAGGeneratorParameters):
	def __init__(self, dataseg, analseg):
		self.dataseg = dataseg
		self.analseg = analseg

	def __repr__(self):
		return repr(self.dataseg)

def DAGGeneratorParameterList(clock, config, available, used):
	# Hack out the out segmentlist we care about.
	available = available.get("available", segments.segmentlist())

	# Leave behind a cusion.
	if len(available):
		if float(clock.now() - available[-1][1]) < config.cusion_timeout:
			available -= segments.segmentlist([segments.segment(available[-1][1] - config.analysis_length_minimum, available[-1][1])])

	# Calculate the unused analysis segments.
	analsegs = available - used.get("analysis", segments.segmentlist())

	# For each analysis segment, identify the science segment in which
	# it occurs.
	scisegs = segments.segmentlist(map(lambda seg: available[available.find(seg)], analsegs))

	# Construct a list of data segment and analysis segment pairs.
	pairs = map(lambda sciseg, analseg: compute_boundaries(config, sciseg, analseg), scisegs, analsegs)

	# Throw out null results.
	pairs = filter(lambda pair: pair != (None, None), pairs)

	# Return a list of DAGGeneratorParameters from the data and
	# analysis segments.
	return map(lambda pair: DAGGeneratorParameters(*pair), pairs)

class DAGGenerator(onasys.DAGGenerator.DAGGenerator):
	def dag_name(self):
		return self.tag + "-" + str(int(self.params.dataseg[0])) + "-" + str(int(self.params.dataseg.duration())) + ".dag"

	def generate_command(self):
		command = self.config.data_dir + "/" + self.config.generator + " -s " + str(self.params.dataseg[0]) + " -e " + str(self.params.dataseg[1]) + " -a " + str(self.params.analseg[0]) + " -b " + str(self.params.analseg[1]) + " -f " + self.dag_name() + " -t " + self.config.data_dir
		if self.config.extra_args:
			command += " " + self.config.extra_args
		return command

	def generate(self):
		logging.info(self.generate_command())
		child = popen2.Popen3(self.generate_command(), True)
		for line in child.childerr:
			logging.warning(line.rstrip())
		newlyused = segments.segmentlist()
		try:
			for line in child.fromchild:
				newlyused |= segments.segmentlist([segments.segment(*map(lal.LIGOTimeGPS, line.split()))])
		except Exception, e:
			raise onasys.DAGGenerator.DAGGeneratorError, "error parsing generator output \"" + line.rstrip() + "\": " + str(e)
		if len(newlyused) != 1:
			logging.warning("generator returned more than one analysis segment for removal")
		if newlyused.extent() not in self.params.analseg:
			logging.warning("analysis segment to remove exceeds bounds of " + str(self.params.analseg))
		status = child.wait()
		if not os.WIFEXITED(status) or os.WEXITSTATUS(status):
			raise onasys.DAGGenerator.DAGGeneratorError, "generator returned " + str(os.WEXITSTATUS(status))
		return {"analysis": newlyused}
